
#content h3 {
    font-family:Lucida Grande,Verdana,Arial,Helvetica,sans serif;
    font-size:13px;
    font-size-adjust:none;
    font-stretch:normal;
    font-style:normal;
    font-variant:normal;
    font-weight:bold;
    line-height:14px;
    margin:10px 0pt 0pt 10px;
    padding:0pt;
    text-transform:lowercase;
}

ul.sortablelist {
    list-style-image:none;
    list-style-type:none;
    margin:0px;
    padding:0px;
}

ul.draggable-container li {
    margin:0px;
    padding:0px;
}

li.draggable-orange {
    background-color:#FFF4D8;
    border:1px solid #E8A400;
}

li.draggable-green {
    background-color:#ECF3E1;
    border:1px solid #C5DEA1;
    cursor:move;
}

span.handle {
    background-color:#E8A400;
    color:white;
    cursor:move;
}

.hoverClass {
    background-color:#FFFFFF;
}

.kss-dragdrop-ghost {
    opacity: 0.6;
    filter:alpha(opacity=60);
}

.kss-dragdrop-clone {
    visibility: hidden;
}

.kss-dragdrop-clone-border {
    border-style: solid;
    border: 1px dotted gray;
}
